from collections import defaultdict
from django.conf import settings
from django.db import transaction
from django.db.models import Q, Prefetch, prefetch_related_objects, Count, Max
from django.utils.functional import cached_property
from rest_framework.exceptions import ValidationError, NotFound
from rest_framework.generics import (
    ListAPIView, CreateAPIView, DestroyAPIView, ListCreateAPIView, RetrieveUpdateDestroyAPIView,
    UpdateAPIView, get_object_or_404
)
from rest_framework import status, serializers
from rest_framework.response import Response
from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import (
    Corpus, Element, ElementPath, Right, MetaData, AllowedMetaData,
    Classification, ClassificationState, Transcription, ElementType
)
from arkindex.documents.serializers.elements import (
    CorpusSerializer, ElementSerializer, ElementSlimSerializer, ElementListSerializer,
    ElementCreateSerializer, ElementNeighborsSerializer, ElementParentSerializer,
    MetaDataUpdateSerializer, ElementTypeSerializer, ElementBulkSerializer,
)
from arkindex.documents.serializers.light import CorpusAllowedMetaDataSerializer, ElementTypeLightSerializer
from arkindex.documents.serializers.ml import ElementTranscriptionSerializer
from arkindex.dataimport.models import WorkerVersion
from arkindex.images.models import Zone
from arkindex.project.openapi import AutoSchema
from arkindex.project.mixins import CorpusACLMixin, SelectionMixin
from arkindex.project.pagination import PageNumberPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly, IsAuthenticated
from arkindex.project.triggers import corpus_delete
from arkindex.project.tools import BulkMap

import operator
import uuid


classifications_queryset = Classification.objects.select_related('ml_class', 'source').order_by('-confidence')

best_classifications_prefetch = Prefetch(
    'classifications',
    queryset=classifications_queryset
    # Keep best or validated distinct classifications
    .filter(Q(state=ClassificationState.Validated) | Q(high_confidence=True))
    # Remove rejected classifications
    .exclude(state=ClassificationState.Rejected).distinct(),
    to_attr='best_classes'
)


def _fetch_children_count(elements):
    """
    From a list of elements, sets the `children_count` attribute to a boolean
    indicating if an element has children elements.

    The initial goal was to have a way for the frontend to tell if an element had any children;
    having a count was optional. It is possible to just use an Exists subquery since Django 2.2:
    Element.objects.annotate(has_children=Exists(ElementPath.objects.filter(path__last=OuterRef('pk'))))
    This works but makes the element list queries 20 times slower, due to the way the Postgres query planner
    handles this.

    Instead, we use a second query only on element paths with a GROUP BY, then re-link the results back to
    Element instances after Django REST Framework does the pagination. This only makes the query twice as slow,
    and this is made optional in a `with_children_count` GET parameter. The difference in execution time
    between telling if an element has children and telling how many children an element has becomes negligible.

    The ElementsListMixin uses arkindex.project.tools.BulkMap to apply this method and perform the second request
    *after* DRF's pagination, because there is no way to perform post-processing after pagination in Django
    without having to use Django private methods.
    """
    counts = dict(
        ElementPath.objects
                   .filter(path__last__in=map(operator.attrgetter('id'), elements))
                   .values('path__last')
                   .annotate(count=Count('element_id', distinct=True))
                   .values_list('path__last', 'count')
    )

    for element in elements:
        element.children_count = counts.get(element.id, 0)

    return elements


class ElementsListMixin(object):
    """
    Provides common features across element list endpoints (list, parents, children)
    """
    serializer_class = ElementListSerializer
    queryset = Element.objects.all()
    openapi_overrides = {
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'type',
                'in': 'query',
                'description': 'Filter elements by type',
                'required': False,
                'schema': {
                    'type': 'string',
                }
            },
            {
                'name': 'name',
                'in': 'query',
                'description': 'Only include elements whose name contains a given string (case-insensitive)',
                'required': False,
                'schema': {'type': 'string'},
            },
            {
                'name': 'best_class',
                'in': 'query',
                'description': 'Restrict to or exclude elements with best classes, '
                               'or restrict to elements with a specific best class',
                'required': False,
                'schema': {'anyOf': [
                    {'type': 'string', 'format': 'uuid'},
                    {'type': 'boolean'},
                ]}
            },
            {
                'name': 'folder',
                'in': 'query',
                'description': 'Restrict to or exclude elements with folder types',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
            {
                'name': 'worker_version',
                'in': 'query',
                'description': 'Only include elements created by a specific worker version',
                'required': False,
                'schema': {
                    'type': 'string',
                    'format': 'uuid',
                }
            },
            {
                'name': 'with_best_classes',
                'in': 'query',
                'description': (
                    'Returns best classifications for each element. '
                    'If not set, elements best_classes field will always be null'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False,
                }
            },
            {
                'name': 'with_children_count',
                'in': 'query',
                'description': (
                    'Include the `children_count` integer to count direct children elements of each element. '
                    'Otherwise, `children_count` will always be null.'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            }
        ]
    }

    @cached_property
    def selected_corpus(self):
        corpus_id = self.request.query_params.get('corpus')
        if corpus_id is None:
            return

        # When a corpus is specified, check independently that it's readable
        # by the current user. This prevents comparing the corpus of every element
        try:
            return self.get_corpus(uuid.UUID(corpus_id))
        except ValueError:
            raise ValidationError({'corpus': ['Not a valid uuid']})

    def get_filters(self):
        filters = {}
        if self.selected_corpus:
            filters['corpus'] = self.selected_corpus
        else:
            filters['corpus__in'] = Corpus.objects.readable(self.request.user)

        if 'name' in self.request.query_params:
            filters['name__icontains'] = self.request.query_params['name']

        if 'type' in self.request.query_params:
            filters['type__slug'] = self.request.query_params['type']

        only_folder = self.request.query_params.get('folder')
        if only_folder is not None:
            filters['type__folder'] = only_folder.lower() not in ('false', '0')

        if 'worker_version' in self.request.query_params:
            try:
                worker_version_id = uuid.UUID(self.request.query_params['worker_version'])
            except (TypeError, ValueError):
                raise ValidationError({'worker_version': ['Invalid UUID']})

            if not WorkerVersion.objects.filter(id=worker_version_id).exists():
                raise ValidationError({'worker_version': ['This worker version does not exist.']})

            filters['worker_version_id'] = worker_version_id

        return filters

    def get_classifications_filters(self):
        """
        Build Django ORM filters using Q expressions related to best classes.
        Supports 3 modes:
        - elements without any best classes
        - elements with any best classes
        - elements with a specific best class
        """
        class_filter = self.request.query_params.get('best_class')
        if class_filter is None:
            return

        # Generic ORM query to find best classes:
        # - elements with a validated classification
        # - OR where high confidence is True
        best_classifications = Q(classifications__state=ClassificationState.Validated) \
            | Q(classifications__high_confidence=True)

        # List elements without any best classes, by inverting the query above
        if class_filter.lower() in ('false', '0'):
            return ~best_classifications

        try:
            # Filter on a specific class
            class_filter = uuid.UUID(class_filter)
            return best_classifications & Q(classifications__ml_class_id=class_filter)
        except (TypeError, ValueError):
            # By default, use all best classifications
            return best_classifications

    def get_prefetch(self):
        prefetch = {'corpus', 'zone__image__server', 'type'}

        with_best_classes = self.request.query_params.get('with_best_classes')
        if with_best_classes and with_best_classes.lower() not in ('false', '0'):
            prefetch.add(best_classifications_prefetch)

        return prefetch

    def get_order_by(self):
        return ('corpus', 'type__slug', 'name', 'id')

    def filter_queryset(self, queryset):
        # Ordering by ID is required by postgres to order elements with common corpus, type and name
        queryset = queryset \
            .filter(**self.get_filters()) \
            .prefetch_related(*self.get_prefetch()) \
            .order_by(*self.get_order_by())

        class_filters = self.get_classifications_filters()
        if class_filters is not None:
            # Use queryset.distinct() whenever best_class is defined
            queryset = queryset.filter(class_filters).distinct()

        with_children_count = self.request.query_params.get('with_children_count')
        if with_children_count and with_children_count.lower() not in ('false', '0'):
            queryset = BulkMap(_fetch_children_count, queryset)

        return queryset


class ElementsList(ElementsListMixin, CorpusACLMixin, ListAPIView):
    """
    List elements and filter by type, name, corpus, ML class
    """
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListElements',
        'description': 'List elements and filter by type, name, corpus, ML class',
        'parameters': ElementsListMixin.openapi_overrides['parameters'] + [
            {
                'name': 'corpus',
                'in': 'query',
                'description': 'Filter elements by corpus ID',
                'required': False,
                'schema': {'type': 'string', 'format': 'uuid'},
            },
            {
                'name': 'top_level',
                'in': 'query',
                'description': 'Only include folder elements without parent elements (top-level elements).',
                'required': False,
                'schema': {'type': 'boolean', 'default': False},
            }
        ]
    })

    @property
    def is_top_level(self):
        return self.request.query_params.get('top_level') not in (None, 'false', '0')

    def get_filters(self):
        filters = super().get_filters()

        if self.is_top_level:
            # Require a corpus
            if not self.selected_corpus:
                raise ValidationError("Missing corpus for top-level listing.")

            # Oly return folders without parents
            filters['paths__isnull'] = True
            filters['type__in'] = self.selected_corpus.types.filter(folder=True)

        return filters

    def get_order_by(self):

        if self.is_top_level:
            return ('name', 'id')

        return super().get_order_by()


class ElementParents(ElementsListMixin, ListAPIView):
    """
    List all parents of an element
    """
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListElementParents',
        'description': 'List all parents of an element',
        'parameters': ElementsListMixin.openapi_overrides['parameters'] + [
            {
                'name': 'recursive',
                'in': 'query',
                'description': 'List recursively (parents, grandparents, etc.)',
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            },
        ]
    })

    @property
    def is_recursive(self):
        recursive_param = self.request.query_params.get('recursive')
        return recursive_param is not None and recursive_param.lower() not in ('false', '0')

    def get_filters(self):
        filters = super().get_filters()

        if not self.is_recursive:
            # List direct parents only: elements whose IDs are the last in the element's paths
            filters['id__in'] = ElementPath.objects \
                .filter(element_id=self.kwargs['pk']) \
                .values('path__last')

        return filters

    def filter_queryset(self, queryset):
        if not self.is_recursive:
            return super().filter_queryset(queryset)

        # Redefine the queryset here as get_ascending needs its filters as an argument
        # and returns a list instead of an actual queryset
        parents = Element.objects.get_ascending(self.kwargs['pk'], **self.get_filters())

        # The list is unsorted; prefetch using prefetch_related_objects and sort with sorted
        if not self.is_recursive:
            # Apply .distinct() using a dict
            parents = list({parent.id: parent for parent in parents}.values())

        prefetch_related_objects(parents, *self.get_prefetch())

        with_children_count = self.request.query_params.get('with_children_count')
        if with_children_count and with_children_count.lower() not in ('false', '0'):
            _fetch_children_count(parents)

        return sorted(parents, key=operator.attrgetter('corpus', 'type.slug', 'name', 'id'))


class ElementChildren(ElementsListMixin, ListAPIView):
    """
    List all children of an element
    """
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListElementChildren',
        'description': 'List all children of an element',
        'parameters': ElementsListMixin.openapi_overrides['parameters'] + [
            {
                'name': 'recursive',
                'in': 'query',
                'description': 'List recursively (children, grandchildren, etc.)',
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            },
        ]
    })

    def get_filters(self):
        filters = super().get_filters()
        recursive_param = self.request.query_params.get('recursive')

        if recursive_param is None or recursive_param.lower() in ('false', '0'):
            # Only list direct children
            # We could use Element.objects.filter(paths__path__last=volid) directly,
            # but it is 100× slower due to the implementation of path__last.
            # get_descending uses 'path @> ARRAY[element.id]', which is much faster,
            # so we append the __last filter to it.
            filters['paths__path__last'] = self.kwargs['pk']

        return filters

    def get_order_by(self):
        return ('paths__ordering', )

    def get_queryset(self):
        return Element.objects.get_descending(self.kwargs['pk'])


class ElementRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Retrieve a single element
    """
    serializer_class = ElementSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['elements'],
    }

    def get_object(self):
        # Prevent duplicating database request
        if not hasattr(self, 'element'):
            self.element = super().get_object()
        return self.element

    def get_queryset(self):
        corpora = Corpus.objects.readable(self.request.user)
        return Element.objects \
            .filter(corpus__in=corpora) \
            .select_related(
                'corpus',
                'type',
                'zone__image__server',
            ) \
            .prefetch_related(
                'metadatas__revision__repo',
                'metadatas__entity',
                Prefetch('classifications', queryset=classifications_queryset)
            )

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        rights = obj.corpus.get_acl_rights(request.user)
        if request.method == 'DELETE' and Right.Admin not in rights:
            self.permission_denied(request, message='You do not have admin access to this element.')
        elif request.method != 'GET' and Right.Write not in rights:
            self.permission_denied(request, message='You do not have write access to this element.')

    def perform_destroy(self, instance):
        children_count = ElementPath.objects.filter(path__contains=[instance.id]).count()
        if children_count:
            raise ValidationError({
                'element': "Element '{}' is linked to {} children elements".format(instance.id, children_count)
            })
        super().perform_destroy(instance)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and self.request.method != 'GET':
            context['element'] = self.get_object()
        return context


class ElementNeighbors(ListAPIView):
    """
    List neighboring elements
    """
    serializer_class = ElementNeighborsSerializer
    openapi_overrides = {
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'n',
                'in': 'query',
                'description': 'Number of neighbors to retrieve around the element',
                'required': False,
                'schema': {
                    'type': 'integer',
                    'minimum': 1,
                    'maximum': 10,
                },
            },
        ]
    }

    def get_queryset(self):
        n = self.request.query_params.get('n', 1)
        try:
            n = int(n)
        except (TypeError, ValueError):
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})
        if not 1 <= n <= 10:
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})

        try:
            element = Element.objects.get(id=self.kwargs['pk'])
        except Element.DoesNotExist:
            raise NotFound

        return Element.objects.get_neighbors(element, n)


class NoRequestBodySchema(AutoSchema):
    def _get_request_body(self, path, method):
        # Remove request body from OpenAPI schema
        return None


class ElementParent(CreateAPIView, DestroyAPIView):
    """
    Manage element relation with one of its parents
    """
    serializer_class = ElementParentSerializer
    permission_classes = (IsVerified, )
    schema = NoRequestBodySchema()
    openapi_overrides = {
        'tags': ['elements'],
    }

    def get_serializer_from_params(self, child=None, parent=None, **kwargs):
        data = {'child': child, 'parent': parent}
        kwargs['context'] = self.get_serializer_context()
        return ElementParentSerializer(data=data, **kwargs)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        # Not necessary to serialize again
        serializer.create(serializer.validated_data)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def destroy(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        serializer.delete(serializer.validated_data)
        return Response(status=status.HTTP_204_NO_CONTENT)


class ManageSelection(SelectionMixin, ListAPIView):
    """
    Manage all selected elements
    """
    serializer_class = ElementListSerializer
    permission_classes = (IsAuthenticated, )
    openapi_overrides = {
        'operationId': 'ManageSelection',
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'with_best_classes',
                'in': 'query',
                'description': (
                    'Returns best classifications for each element. '
                    'If not set, elements best_classes field will always be null'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            }
        ]
    }

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        if not settings.ARKINDEX_FEATURES['selection']:
            raise ValidationError(['Selection is not available on this instance.'])

    def get_queryset(self):
        filtered_queryset = self.get_selection().order_by('corpus', 'type', 'name')

        with_best_classes = self.request.query_params.get('with_best_classes')
        if with_best_classes and with_best_classes.lower() not in ('false', '0'):
            return filtered_queryset.prefetch_related(best_classifications_prefetch)

        return filtered_queryset

    def post(self, request, *args, **kwargs):
        if 'ids' not in request.data:
            raise ValidationError({'ids': ['Element IDs are required.']})

        # Strange field validation done here,
        # as we have different serializers on each request method on the same endpoint.
        # Using serializer fields means we still preserve DRF's normal validation.
        field = serializers.ListField(child=serializers.UUIDField())
        element_ids = field.to_internal_value(request.data['ids'])

        elements = Element.objects.filter(
            id__in=element_ids,
            corpus__in=Corpus.objects.readable(request.user)
        )

        if elements.count() != len(element_ids):
            raise ValidationError({'ids': [
                'Some element IDs do not exist or you do not have permission to access them.'
            ]})

        # Add new Selection objects in bulk, ignoring elements that were already selected
        self.request.user.selected_elements.add(
            *elements.exclude(selection_users=request.user).values_list('id', flat=True)
        )

        return Response(
            status=status.HTTP_201_CREATED,
            data=ElementSlimSerializer(
                elements.prefetch_related('corpus', 'zone__image__server', 'type'),
                context={'request': request},
                many=True
            ).data,
        )

    def delete(self, request, *args, **kwargs):
        if request.data.get('id'):
            field = serializers.PrimaryKeyRelatedField(queryset=self.get_selection())
            element = field.to_internal_value(request.data['id'])
            request.user.selections.get(element=element).delete()
        else:
            request.user.selections.all().delete()

        return Response(status=status.HTTP_204_NO_CONTENT)


class CorpusList(ListCreateAPIView):
    """
    List all available corpora and create new ones
    """
    serializer_class = CorpusSerializer
    pagination_class = None
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['corpora'],
    }

    def get_queryset(self):
        return Corpus.objects \
                     .readable(self.request.user, add_stats=True) \
                     .prefetch_related('types')


class CorpusRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Get, update and delete corpora
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['corpora'],
    }

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        rights = obj.get_acl_rights(request.user)
        if request.method == 'DELETE' and Right.Admin not in rights:
            self.permission_denied(request, message='You do not have admin acces to this corpus.')
        elif request.method != 'GET' and Right.Write not in rights:
            self.permission_denied(request, message='You do not have write access to this corpus.')

    def perform_destroy(self, instance):
        corpus_delete(instance)


class TranscriptionsPagination(PageNumberPagination):
    page_size = 100


class ElementTranscriptions(ListAPIView):
    """
    List all transcriptions for an element, optionally filtered by type or worker version id.
    Recursive parameter allow listing transcriptions on sub-elements,
    otherwise element fields in the response will be set to null.
    """
    serializer_class = ElementTranscriptionSerializer
    pagination_class = TranscriptionsPagination
    openapi_overrides = {
        'operationId': 'ListTranscriptions',
        'security': [],
        'tags': ['transcriptions'],
        'parameters': [
            {
                'name': 'type',
                'in': 'query',
                'required': False,
                'description': 'Transcription type filter',
                'schema': {
                    'type': 'string',
                    'enum': [ts_type.value for ts_type in TranscriptionType],
                }
            }, {
                'name': 'recursive',
                'in': 'query',
                'required': False,
                'description': 'Recursively list transcriptions on sub-elements',
                'schema': {'type': 'boolean'}
            }, {
                'name': 'worker_version',
                'in': 'query',
                'required': False,
                'description': 'Filter transcriptions by worker version',
                'schema': {'type': 'string', 'format': 'uuid'}
            },
        ]
    }

    @property
    def is_recursive(self):
        if not self.request:
            return
        recursive = self.request.query_params.get('recursive')
        return recursive is not None and recursive.lower() not in ('false', '0')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Do serialize the element attached to each transcription in recursive mode only
        context['ignore_element'] = not self.is_recursive
        return context

    def check_object_permissions(self, request, element):
        super().check_object_permissions(request, element)
        if element.type.folder:
            self.permission_denied(request, message='Element is a folder')

    def get_queryset(self):
        element = get_object_or_404(Element.objects.filter(
            id=self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user)
        ))
        self.check_object_permissions(self.request, element)

        queryset = Transcription.objects \
            .prefetch_related('zone__image__server', 'source') \
            .extra(
                # ORDER BY casting IDs as char to avoid PostgreSQL optimizer inefficient scan
                select={'char_id': 'CAST(id AS CHAR(36))'},
                order_by=['char_id']
            )

        if self.is_recursive:
            queryset = queryset.filter(
                # Retrieve both element and sub-elements transcriptions
                element__in=[element.id, *Element.objects.get_descending(element.id).values_list('id')]
            ).prefetch_related('element__type', 'element__zone')
        else:
            queryset = queryset.filter(element_id=element.id)

        return queryset

    def filter_queryset(self, queryset):
        # Filter by transcription type
        req_type = self.request.query_params.get('type')
        if req_type:
            try:
                req_type = TranscriptionType(req_type)
                queryset = queryset.filter(type=req_type)
            except ValueError:
                raise ValidationError({'type': 'Not a valid transcription type'})

        # Filter by worker version
        worker_version = self.request.query_params.get('worker_version')
        if worker_version:
            queryset = queryset.filter(worker_version_id=worker_version)

        return queryset


class ElementsCreate(CreateAPIView):
    """
    Create a new element
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementCreateSerializer
    openapi_overrides = {
        'operationId': 'CreateElement',
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'slim_output',
                'in': 'query',
                'required': False,
                'description': 'When enabled, the endpoint will only return the new Element\'s ID. This is especially useful for an automated tool, as it speeds up the element creation.',
                'schema': {
                    'type': 'boolean',
                }
            },
        ]
    }


class ElementMetadata(CreateAPIView):
    """
    Create a metadata on an existing element.

    The `date` type allows dates to be parsed and indexed for search APIs. The supported date formats are as follows:

    * `YYYY`: `1995` (recommended)
    * `YYYY-MM`: `1995-02` (recommended)
    * `YYYY-MM-DD`: `1995-02-16` (recommended)
    * `/YYYY(.*)month/`: `1995, february`, with months in French or English (case-insensitive)
    * `YYYY-YYYY`: `1995-1996` (yields a date interval)

    The `reference` type is also indexed in search APIs and allows searching by
    known identifiers or references, or any arbitrary string.
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataUpdateSerializer
    openapi_overrides = {
        'operationId': 'CreateMetaData',
        'tags': ['elements']
    }

    def get_queryset(self):
        # We need to use the default database to avoid stale read
        # when a metadata is created immediately after an element creation
        return Element.objects.using('default').filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:  # Ignore this step when generating the schema with OpenAPI
            context['element'] = self.get_object()
        return context


class MetadataEdit(RetrieveUpdateDestroyAPIView):
    """
    Edit an existing element medadata
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataUpdateSerializer
    openapi_overrides = {
        'tags': ['elements']
    }

    def get_queryset(self):
        # Filter readable metadata in order to check object permissions
        return MetaData.objects.filter(element__corpus__in=Corpus.objects.readable(self.request.user))

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if Right.Write not in obj.element.corpus.get_acl_rights(request.user):
            self.permission_denied(request, message='You do not have write access to this corpus.')

    def perform_destroy(self, instance):
        serializer = self.get_serializer(instance)
        serializer.delete(instance)


class AllowedMetaDataPagination(PageNumberPagination):
    page_size = 200


class CorpusAllowedMetaData(CorpusACLMixin, ListAPIView):
    """
    List allowed metadata (type, name) couples in a corpus
    """
    serializer_class = CorpusAllowedMetaDataSerializer
    pagination_class = AllowedMetaDataPagination
    openapi_overrides = {
        'operationId': 'ListCorpusAllowedMetaDatas',
        'description': 'List (type, name) couples allowed for corpus metadata',
        'tags': ['corpora']
    }

    def get_queryset(self):
        if not self.request:
            return AllowedMetaData.objects.none()
        return AllowedMetaData.objects.filter(corpus=self.get_corpus(self.kwargs['pk']))


class ElementTypeCreate(CreateAPIView):
    """
    Create a new element type for a specific corpus
    """
    serializer_class = ElementTypeSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['elements']
    }


class ElementTypeUpdate(UpdateAPIView):
    """
    Update an existing element type
    """
    serializer_class = ElementTypeLightSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'operationId': 'UpdateElementType',
        'tags': ['elements']
    }

    def get_queryset(self):
        return ElementType.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))


class ElementBulkCreate(CreateAPIView):
    """
    Create multiple child elements at once on a single parent
    """
    serializer_class = ElementBulkSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'CreateElements',
        'tags': ['elements'],
        'responses': {
            '201': {
                'description': 'Created',
                'content': {
                    'application/json': {
                        'examples': [{
                            'value': [{
                                'id': 'string'
                            }]
                        }]
                    }
                }
            }
        }
    }

    def get_object(self):
        if not hasattr(self, 'element'):
            self.element = super().get_object()
            if not self.element.zone_id:
                raise ValidationError({'element': ['Cannot create child elements on an element without a zone.']})
        return self.element

    def get_queryset(self):
        if not self.request:
            return Element.objects.none()
        return Element \
            .objects \
            .filter(corpus__in=Corpus.objects.writable(self.request.user)) \
            .select_related('zone') \
            .only('id', 'corpus_id', 'zone__id', 'zone__image_id')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context['element'] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = self.perform_create(serializer)
        headers = self.get_success_headers({})
        return Response(data, status=status.HTTP_201_CREATED, headers=headers)

    @transaction.atomic
    def perform_create(self, serializer):
        worker_version = serializer.validated_data.get('worker_version')
        elements = serializer.validated_data['elements']

        # Retrieve existing zones from the DB
        image_id = self.element.zone.image_id
        existing_zones = {
            # Use WKB represensation to compare existing zones
            # to avoid comparing references or slower coordinates
            polygon.wkb: zone_id
            for polygon, zone_id in Zone.objects.filter(image_id=image_id).values_list('polygon', 'id')
        }

        # Retrieve or create required zones
        new_zones = []
        for element_data in elements:
            polygon = element_data.pop('polygon')
            # Retrieve or create the zone on the image
            zone_id = existing_zones.get(polygon.wkb)
            if not zone_id:
                new_zone = Zone(
                    id=uuid.uuid4(),
                    image_id=image_id,
                    polygon=polygon,
                )
                zone_id = new_zone.id
                existing_zones[polygon.wkb] = zone_id
                new_zones.append(new_zone)
            element_data['zone'] = zone_id

        Zone.objects.bulk_create(new_zones)

        # ElementPath.get_next_order, but for multiple types
        next_orderings = defaultdict(
            int,
            ElementPath
            .objects
            .filter(
                path__last=self.element.id,
                element__type_id__in=set(element_data['type'] for element_data in elements)
            )
            .values('element__type_id')
            .annotate(next_ordering=Max('ordering') + 1)
            .values_list('element__type_id', 'next_ordering')
        )
        # Prepare elements
        for element_data in elements:
            # Keep ordering in the element payload to optimize element path creation later
            element_data['ordering'] = next_orderings[element_data['type']]
            next_orderings[element_data['type']] += 1

            element_data['element'] = Element(
                id=uuid.uuid4(),
                zone_id=element_data['zone'],
                corpus_id=self.element.corpus_id,
                type_id=element_data['type'],
                name=element_data['name'],
                worker_version=worker_version,
            )

        Element.objects.bulk_create([
            element_data['element'] for element_data in elements
        ])
        ElementPath.objects.bulk_create([
            ElementPath(
                path=parent_path + [self.element.id],
                element_id=element_data['element'].id,
                ordering=element_data['ordering'],
            )
            for parent_path in self.element.paths.all().values_list('path', flat=True)
            for element_data in elements
        ])

        return [{'id': element_data['element'].id} for element_data in elements]
