from django.db import models
from itertools import groupby, chain
import uuid


class ElementManager(models.Manager):
    """Model manager for elements"""

    def get_ascending_paths(self, child_id, **filters):
        """
        Get all parent paths for a specific element ID.
        """
        # TODO: it should be possible to do in a single query through Django's filtered relations
        from arkindex.documents.models import ElementPath
        paths = ElementPath.objects.filter(element_id=child_id).values_list('path', flat=True)
        parents = {
            parent.id: parent
            for parent in self.filter(id__in=chain(*paths), **filters)
        }
        return [
            filter(None, [
                parents.get(parent_id)
                for parent_id in path
            ])
            for path in paths
        ]

    def get_ascending(self, child_id, **filters):
        """
        Get all parent elements for a specific element ID.
        > All the elements from paths
        """
        return list(chain(*self.get_ascending_paths(child_id, **filters)))

    def get_ascendings_paths(self, *children_ids, **filters):
        """
        Get all ascending paths for some elements IDs.
        """
        # Loads paths and group them by element ids
        from arkindex.documents.models import ElementPath
        paths = ElementPath.objects.filter(element_id__in=children_ids).order_by('element_id')
        tree = {
            elt_id: [p.path for p in elt_paths]
            for elt_id, elt_paths in groupby(paths, lambda e: e.element_id)
        }

        # Load parents elements in bulk
        parents = {
            parent.id: parent
            for parent in self.filter(id__in=chain(*[p.path for p in paths]), **filters)
        }
        return {
            elt_id: [
                [
                    parents[parent_id]
                    for parent_id in path
                ]
                for path in paths
            ]
            for elt_id, paths in tree.items()
        }

    def get_ascendings(self, *children_ids, **filters):
        """
        Get all parent elements for some element IDs.
        """
        # Just assemble the paths together
        return {
            elt_id: chain(*paths)
            for elt_id, paths in self.get_ascendings_paths(*children_ids, **filters).items()
        }

    def get_descending(self, parent_id, **filters):
        """
        Get all child elements for a specific element ID.
        > All the elements where parent is in the path
        """
        return self.filter(paths__path__contains=[parent_id], **filters).order_by('paths__ordering').distinct()

    def get_descendings(self, parents_ids, prefetch=(), **filters):
        """
        Get all child elements for some element IDs.
        """
        assert all(isinstance(x, uuid.UUID) for x in parents_ids), 'UUIDs are required'

        # Loads paths and group them by element ids
        # The map to str is needed to avoid a cast error in PG
        from arkindex.documents.models import ElementPath
        paths = ElementPath.objects.filter(path__overlap=map(str, parents_ids)).order_by('element_id', 'ordering')
        tree = {
            parent_id: [
                p.element_id
                for p in paths
                if parent_id in p.path
            ]
            for parent_id in parents_ids
        }

        # Load children elements in bulk
        children = self.filter(id__in=[p.element_id for p in paths], **filters)
        if prefetch:
            children = children.prefetch_related(*prefetch)
        return {
            parent_id: [
                child
                for child in children
                if child.id in tree.get(parent_id)
            ]
            for parent_id in parents_ids
        }

    def get_neighbor_paths(self, element, n=1):
        """
        Get n neighbor paths: current, previous and next element paths

        This uses a single raw SQL query to use row numbers, instead of the ordering numeric values,
        to avoid issues with duplicated or missing orderings.

        * The `neighbors` subquery lists all paths for an element, and then list all possible neighbors
          with the same parents. An ordering consistent with rows numbers is added as `position`
        * The main query combines the `neighbors` subquery with itself to have a full product, and only retains
          the combinations with a distance under the requested limit, with elements of the same type
        * The output is ordered by parent path, then by position (row numbers)
        """
        from arkindex.documents.models import ElementPath
        return ElementPath.objects.raw(
            """
            with neighbors as (
              select link.*, row_number() over (partition by p.id order by link.ordering) as position
              from documents_elementpath p
              inner join documents_elementpath link on (p.path=link.path)
              inner join documents_element as e on (e.id=link.element_id and e.type_id = %(element_type)s)
              where p.element_id = %(element_id)s
            )

            select distinct n.*
            from neighbors as n
            inner join neighbors as m on (m.element_id = %(element_id)s and m.path = n.path)
            where abs(n.position - m.position) <= %(n)s
            order by n.path, n.position, n.element_id
            """,
            params={
                'element_id': element.id,
                'element_type': element.type_id,
                'n': n,
            }
        )

    def get_neighbors(self, element, n=1):
        """
        Returns a list of neighboring ElementPaths for an element, with a prefetched `element` attribute
        and a list of prefetched parent elements in the `parents` attribute.
        """
        paths = list(self.get_neighbor_paths(element, n))
        models.prefetch_related_objects(paths, 'element__type')
        # Build a set of all IDs to load from all paths, then load them into a dict
        elements = {
            elt.id: elt
            for elt in self.filter(id__in=set(chain(*(p.path for p in paths)))).select_related('type')
        }
        # Dispatch the elements by ID to each path
        for path in paths:
            path.parents = list(map(elements.get, path.path))
        return paths


class CorpusManager(models.Manager):
    '''
    Add ACL functions to corpus listing
    '''

    def readable(self, user, add_stats=False):
        qs = super().get_queryset().order_by('name')

        # add_stats is used to determine whether to return the count of users with rights on the corpora
        # concerned by the request. The count of users is made here because the application of certain
        # permissions made in the continuation of this function restricts the access to corpus rights which
        # does not concern the user if he is not an administrator
        if add_stats:
            qs = qs.annotate(authorized_users=models.Count('corpus_right'))

        # An anonymous has access to public corpus only
        if user.is_anonymous:
            return qs.filter(public=True)

        # Admins and internal users have access to every corpus
        if user.is_admin or user.is_internal:
            return qs.all()

        # An authenticated user has access to:
        # * public corpus
        # * private with ACL
        qs = qs.prefetch_related('corpus_right')
        qs = qs.filter(models.Q(public=True) | models.Q(corpus_right__user=user)).distinct()
        return qs

    def writable(self, user):
        # An anonymous user cannot write anything
        if user.is_anonymous:
            return super().none()

        qs = super().get_queryset().order_by('name')

        # Admins and internal users have access to every corpus
        if user.is_admin or user.is_internal:
            return qs.all()

        # Authenticated users can write only on corpora with ACL
        return qs.filter(corpus_right__user=user, corpus_right__can_write=True).distinct()

    def admin(self, user):
        # An anonymous user cannot manage anything
        if user.is_anonymous:
            return super().none()

        qs = super().get_queryset().order_by('name')

        # Admins and internal users have access to every corpus
        if user.is_admin or user.is_internal:
            return qs.all()

        # Authenticated users can manage only corpora with ACL
        return qs.filter(corpus_right__user=user, corpus_right__can_admin=True).distinct()
