from django.urls import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex.documents.models import Element
from arkindex.dataimport.models import Worker, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
import uuid


class TestChildrenElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.element = Element.objects.get(name='Volume 1, page 2r')

    def test_element_children(self):
        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.element.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(response.json()['results'], [])

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 8)
        self.assertSetEqual(
            {r['type'] for r in response.json()['results']},
            {'act', 'page'},
        )

        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?recursive&type=surface'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertCountEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Surface A',
                'Surface B',
                'Surface C',
                'Surface D',
                'Surface E',
                'Surface F'
            ],
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?recursive')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(len(response.json()['results']), 15)
        self.assertSetEqual(
            {r['type'] for r in response.json()['results']},
            {'act', 'page', 'surface', 'text_line'},
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?type=page')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Volume 1, page 1r',
                'Volume 1, page 1v',
                'Volume 1, page 2r'
            ]
        )

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?name=2')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Act 2',
                'Volume 1, page 2r',
            ]
        )

        nested_volume = self.corpus.elements.create(
            type=self.vol.type,
            name='Nested volume',
        )
        nested_volume.add_parent(self.vol)

        response = self.client.get(reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}) + '?folder')
        self.assertListEqual(response.json()['results'], [
            {
                "id": str(nested_volume.id),
                "name": nested_volume.name,
                "type": nested_volume.type.slug,
                "zone": None,
                "corpus": {
                    "id": str(self.corpus.id),
                    "name": self.corpus.name,
                    "public": self.corpus.public,
                },
                "best_classes": None,
                "children_count": None,
                "thumbnail_url": nested_volume.thumbnail.s3_url,
                "thumbnail_put_url": None,
            }
        ])

    def test_element_children_worker_version(self):
        worker = Worker.objects.create(
            repository=self.user.credentials.get().repos.get(),
            name='Worker 1',
            slug='worker_1',
            type=MLToolType.Classifier
        )
        version = WorkerVersion.objects.create(
            worker=worker,
            revision=worker.repository.revisions.get(),
            configuration={"test": "test1"}
        )
        self.corpus.elements.filter(name__contains='page 1r').update(worker_version=version)

        with self.assertNumQueries(8):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'worker_version': str(version.id)}
            )

        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            [
                'Volume 1, page 1r',
            ]
        )

    def test_element_children_worker_version_validation(self):
        with self.assertNumQueries(0):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'worker_version': 'blah'},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'worker_version': uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_element_children_ordering(self):
        expected = [
            'Act 4',
            'Act 5',
            'Act 1',
            'Act 2',
            'Act 3'
        ]
        for ordering, name in enumerate(expected):
            Element.objects.get(type__slug='act', name=name).paths.update(ordering=ordering)

        response = self.client.get(
            reverse('api:elements-children', kwargs={'pk': str(self.vol.id)})
            + '?recursive&type=act'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [r['name'] for r in response.json()['results']],
            expected,
        )

    def test_children_with_children_count(self):
        with self.assertNumQueries(9):
            response = self.client.get(
                reverse('api:elements-children', kwargs={'pk': str(self.vol.id)}),
                data={'with_children_count': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(
            {element['name']: element['children_count'] for element in response.json()['results']},
            {
                'Act 1': 1,
                'Act 2': 2,
                'Act 3': 1,
                'Act 4': 1,
                'Act 5': 1,
                'Volume 1, page 1r': 1,
                'Volume 1, page 1v': 0,
                'Volume 1, page 2r': 0,
            }
        )
