from django.urls import reverse
from django.db.models.sql.constants import LOUTER
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex.documents.models import Corpus
from arkindex.dataimport.models import Worker, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
import uuid


class TestListElements(FixtureAPITestCase):

    def test_list_elements_filter_name(self):
        nameSelected = 'ol'
        response = self.client.get(reverse('api:elements'), data={'name': nameSelected})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertIn(nameSelected, result['name'])

    def test_list_elements_wrong_corpus_filter(self):
        nameSelected = 'not_a_corpus_id'
        response = self.client.get(reverse('api:elements'), data={'corpus': nameSelected})
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'corpus': ['Not a valid uuid']}
        )

    def test_list_elements_corpus_uuid4(self):
        """
        Ensure the corpus filter allows any UUID, not just version 4 UUIDs.
        Using the version=x argument on uuid.UUID causes the version and variant bits
        of a UUID to be changed on instanciation; since we generate v4 RFC 4122-compliant UUIDs,
        this would work in most cases - but we could have any UUID version or variant,
        generated by some other tool, which would cause strange 404 errors.
        """
        str_id = '12345678-1234-5678-1234-567812345678'
        self.assertNotEqual(
            uuid.UUID(str_id),
            uuid.UUID(str_id, version=4),
        )
        corpus = Corpus.objects.create(id=uuid.UUID(str_id), name='A corpus', public=True)
        element_type = corpus.types.create(slug='element', display_name='Generic element')
        element = corpus.elements.create(type=element_type, name='Something')

        response = self.client.get(reverse('api:elements'), data={'corpus': str_id})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(element.id))
        self.assertEqual(data['results'][0]['corpus']['id'], str_id)

    def test_list_elements_filter_type(self):
        response = self.client.get(reverse('api:elements'), data={'type': 'act'})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertEqual(result['type'], 'act')

    def test_list_with_children_count(self):
        with self.assertNumQueries(9):
            response = self.client.get(reverse('api:elements'), data={'with_children_count': True})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIsNone(data['next'])

        self.assertDictEqual(
            {element['name']: element['children_count'] for element in data['results']},
            {
                'Act 1': 1,
                'Act 2': 2,
                'Act 3': 1,
                'Act 4': 1,
                'Act 5': 1,
                'Text line': 0,
                'Surface A': 0,
                'Surface B': 0,
                'Surface C': 0,
                'Surface D': 0,
                'Surface E': 0,
                'Surface F': 0,
                'Volume 1': 8,
                'Volume 1, page 1r': 1,
                'Volume 1, page 1v': 0,
                'Volume 1, page 2r': 0,
                'Volume 2': 3,
                'Volume 2, page 1r': 0,
                'Volume 2, page 1v': 0,
                'Volume 2, page 2r': 0,
            }
        )

    def test_list_top_level_left_join(self):
        """
        Ensure the top_level option on ListElements triggers a left join.
        """
        query = self.corpus.elements.filter(paths__isnull=True, name='something').query
        # Ensure the documents_elementpath table is joined via a LEFT JOIN
        self.assertEqual(query.alias_map['documents_elementpath'].join_type, LOUTER)
        # Ensure the IS NULL goes last
        self.assertTrue(str(query).endswith('AND "documents_elementpath"."id" IS NULL)'))

    def test_list_top_level(self):
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:elements',),
                data={'top_level': True, 'corpus': self.corpus.id},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 1', 'Volume 2'],
        )

    def test_list_elements_filter_worker_version(self):
        worker = Worker.objects.create(
            repository=self.user.credentials.get().repos.get(),
            name='Worker 1',
            slug='worker_1',
            type=MLToolType.Classifier
        )
        version = WorkerVersion.objects.create(
            worker=worker,
            revision=worker.repository.revisions.get(),
            configuration={"test": "test1"}
        )
        self.corpus.elements.filter(name__contains='Volume 2').update(worker_version=version)

        with self.assertNumQueries(9):
            response = self.client.get(
                reverse('api:elements'),
                data={'corpus': self.corpus.id, 'worker_version': str(version.id)},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 2, page 1r', 'Volume 2, page 1v', 'Volume 2, page 2r', 'Volume 2'],
        )

    def test_list_elements_filter_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:elements'),
                data={'corpus': self.corpus.id, 'worker_version': 'blah'},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:elements'),
                data={'corpus': self.corpus.id, 'worker_version': uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})
